/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.servlet.bridge;

import java.io.*;
import java.util.HashMap;

import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.AttributesImpl;
import org.xquark.bridge.Mapper;
import org.xquark.bridge.Mapping;
import org.xquark.bridge.XQBridge;
import org.xquark.servlet.util.FileExtensionFilter;
import org.xquark.servlet.util.PostDataParser;
import org.xquark.serialize.XMLSerializer;
import org.xquark.xml.xdbc.XMLConnection;
import org.xquark.xml.xdbc.XMLResultSet;
import org.xquark.xml.xdbc.XMLStatement;

public class XQBridgeServlet extends HttpServlet {

    private static String urlprop = "org.xquark.xdbc.url";
    private static String userprop = "org.xquark.xdbc.user";
    private static String passwordprop = "org.xquark.xdbc.password";

    private XQBridgePool pool;
    private SAXParserFactory factory;
    private HashMap mappings;

    public void init(ServletConfig config) throws ServletException {
        super.init(config);
        factory = SAXParserFactory.newInstance();
        factory.setNamespaceAware(true);
        pool =
            XQBridgePool.getXQBridgePool(
                config.getInitParameter(urlprop),
                config.getInitParameter(userprop),
                config.getInitParameter(passwordprop));
        mappings =
            loadMappingsFromFiles(
                config.getServletContext().getRealPath("/WEB-INF/mappings"));
    }

    public void doGet(HttpServletRequest request, HttpServletResponse response)
        throws ServletException, IOException {
        PrintWriter out = response.getWriter();
        XQBridge bridge = null;
        try {
            bridge = getXQBridge();
            if (bridge == null) {
                reportError(response, "Could not connect to database", "");
            }
            String query = request.getParameter("exec");
            if (query != null) {
                execQuery(bridge.getXMLConnection(), query, out);
            }
            else {
                reportError(
                    response,
                    "Unrecognized command: <B>"
                        + request.getContextPath()
                        + request.getServletPath()
                        + request.getQueryString()
                        + "</B>.",
                    "query.jsp");
                return;
            }
        }
        catch (Exception e) {
            reportException(response, e, "query.jsp");
        }
        finally {
            out.close();
            freeXQBridge(bridge);
        }
    }

    public void doPost(
        HttpServletRequest request,
        HttpServletResponse response)
        throws ServletException, IOException {
        PrintWriter out = response.getWriter();
        String mapping = null;
        String file = null;
        XQBridge bridge = null;
        if (request.getContentType().startsWith("multipart/form-data;")) {
            PostDataParser poster = new PostDataParser(request);
            mapping = poster.getParameter("mapping");
            file = poster.getFileContent("file");
        }
        else if (
            request.getContentType().equals(
                "application/x-www-form-urlencoded")) {
            mapping = request.getParameter("mapping");
            file = request.getParameter("file");
        }
        else {
            reportError(
                response,
                "Unknown encryption type: " + request.getContentType(),
                "insert.jsp");
            return;
        }
        if (mapping == null || file == null) {
            reportError(
                response,
                "Missing parameters for file insertion",
                "insert.jsp");
        }
        else {
            try {
                bridge = pool.getXQBridge();
                String mapFile = (String) mappings.get(mapping);
                Mapping map = bridge.getMapping(new InputSource(mapFile));
                Mapper mapper = map.getMapper();
                mapper.setDocumentBase("file:/c:/");
                /*
                if (file.indexOf("<?xml") > 0)
                  file = file.substring(file.indexOf("<?xml"));
                XMLReader reader = factory.newSAXParser().getXMLReader();
                reader.setContentHandler(mapper);
                reader.parse(new InputSource(new StringReader(file)));
                */
                mapper.insertDocument(file);
                mapper.close();
                map.close();
                reportMessage(
                    response,
                    "Document inserted successfully.",
                    "insert.jsp");
            }
            catch (SAXException ex) {
                if (ex.getException() != null)
                    reportException(response, ex.getException(), "insert.jsp");
                else
                    reportException(response, ex, "insert.jsp");
            }
            catch (Exception ex) {
                reportException(response, ex, "insert.jsp");
            }
            finally {
                freeXQBridge(bridge);
            }
        }
    }

    private XQBridge getXQBridge() {
        try {
            return pool.getXQBridge();
        }
        catch (Exception ex) {
            log("Cannot connect to database", ex);
            return null;
        }
    }

    private void freeXQBridge(XQBridge bridge) {
        try {
            pool.freeXQBridge(bridge);
        }
        catch (Exception ex) {
            log("An error occured while closing connection", ex);
        }
    }

    private HashMap loadMappingsFromFiles(String realPath) {
        HashMap result = new HashMap();
        XQBridge bridge = getXQBridge();
        FileFilter mapFilter = new FileExtensionFilter(".map");
        File mappingsDir = new File(realPath);
        File[] mapDirs = mappingsDir.listFiles();
        for (int i = 0; i < mapDirs.length; i++) {
            File mapDir = mapDirs[i];
            if (mapDir.isDirectory()) {
                File[] mapFiles = mapDir.listFiles(mapFilter);
                if (mapFiles.length > 0) {
                    String mapFile = null;
                    try {
                        mapFile = mapFiles[0].toURL().toString();
                        Mapping mapping =
                            bridge.getMapping(new InputSource(mapFile));
                        result.put(mapDir.getName(), mapFile);
                    }
                    catch (Exception ex) {
                        log("Error while loading mapping file: " + mapFile, ex);
                    }
                }
            }
        }
        return result;
    }

    private void execQuery(XMLConnection conn, String query, PrintWriter out)
        throws Exception {
        XMLStatement stmt = null;
        XMLResultSet xrs = null;
        try {
            stmt = conn.createStatement();
            xrs = stmt.executeQuery(query);
            XMLSerializer serializer = new XMLSerializer(out, "iso-8859-1");
            xrs.setContentHandler(serializer);
            xrs.setLexicalHandler(serializer);
            serializer.startDocument();
            //serializer.processingInstruction("xml-stylesheet", "href=\"xsl/beautify.xsl\" type=\"text/xsl\"");
            serializer.startElement(
                "",
                "results",
                "results",
                new AttributesImpl());
            while (xrs.hasNext())
                xrs.nextAsSAX();
            serializer.endElement("", "results", "results");
            serializer.endDocument();
        }
        finally {
            if (xrs != null)
                xrs.close();
            if (stmt != null)
                stmt.close();
        }
    }

    private void reportMessage(
        HttpServletResponse response,
        String message,
        String url)
        throws IOException {
        response.sendRedirect(
            response.encodeRedirectURL(
                "info.jsp?url=" + url + "&message=" + message));
    }

    private void reportError(
        HttpServletResponse response,
        String error,
        String url)
        throws IOException {
        response.sendRedirect(
            response.encodeRedirectURL(
                "error.jsp?url=" + url + "&message=" + error));
    }

    private void reportException(
        HttpServletResponse response,
        Exception e,
        String url)
        throws IOException {
        /*
        StringWriter sw = new StringWriter();
        e.printStackTrace(new PrintWriter(sw));
        String message = "<pre>"+sw.toString()+"</pre>";
        response.sendRedirect(response.encodeRedirectURL("exception.jsp?url="+url+"&message="+message));
        */
        PrintWriter out = response.getWriter();
        out.println("<html><body><pre>");
        e.printStackTrace(out);
        out.println("</pre></body></html>");
    }
}
